import sys

import numpy as np
import pandas as pd
import os

import warnings
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.model_selection import learning_curve
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report, confusion_matrix, precision_recall_curve

from utilities import plot_precision_recall_curve, plot_learning_curve, import_data, plot_roc_curve, plot_counts
from utilities import ordinal_encode, heat_map, plot_features, plot_confusion_matrix, normalize, print_high_confidence_samples

sys.stdout.reconfigure(line_buffering=True)
warnings.filterwarnings("ignore")

# Constants
y_data = 'class'
y_values = ['normal', 'anomaly']
df_train, df_test, model_name = import_data(
	train_file_path = os.path.join(os.path.dirname(__file__), "nsl-kdd-dataset/" + "KDDTrain+.arff"),
	test_file_path = os.path.join(os.path.dirname(__file__), "nsl-kdd-dataset/" + "KDDTest+.arff"),
	model_name = "Random Forest")

# Defining scalers
sc = StandardScaler()
enc = LabelEncoder()

numerical_columns = df_train.select_dtypes(include=np.number).columns
label_columns = df_train.select_dtypes(include=object, exclude=np.number).columns

# Normalize data
ordinal_encode(df = df_train, categories = y_values, target = y_data)
ordinal_encode(df = df_test, categories = y_values, target = y_data)

normalize(df_train, df_test, y_data, sc, enc)

# Separate X and y
X_train = df_train.select_dtypes(include=[np.number]).drop(columns = [y_data])
X_test = df_test.select_dtypes(include=[np.number]).drop(columns = [y_data])
y_train = df_train[[y_data]]
y_test = df_test[[y_data]]

model = RandomForestClassifier(n_estimators=100, max_depth=None, random_state=42)

def predict(prediction_input):
	if len(prediction_input) == 0:
		return
	input_df = pd.DataFrame(prediction_input, columns=X_train.columns)
	input_df[numerical_columns] = sc.transform(input_df[numerical_columns])
	return ["anomaly" if x == 1 else "normal" for x in model.predict(input_df)]

def train():
	model.fit(X_train, y_train.values.ravel())
	print("Training complete.")

def graphs():
	# Correlation
	heat_map(df_train, model_name=model_name)

	# Predictions
	y_prediction = model.predict(X_test)

	# Plot Confusion Matrix
	plot_confusion_matrix(confusion_matrix=confusion_matrix(y_test, y_prediction),
	                      accuracy=model.score(X_test, y_test),
	                      model_name=model_name)

	print("Classification Report: \n", classification_report(y_test, y_prediction))

	# Get high confidence samples for which the model is 90% confident
	print_high_confidence_samples(model, X_train)

	# Feature Importance Plot
	features = pd.DataFrame(model.feature_importances_, index=X_train.columns, columns=['Importance']).sort_values(
		by='Importance', ascending=False)
	plot_features(features, "Higher importance = More impact on classification", model_name=model_name)

	# Precision-Recall Curve
	print("Calculating Precision Recall Curve")
	y_scores = model.predict_proba(X_test)[:, 1]
	precision, recall, _ = precision_recall_curve(y_test, y_scores)
	plot_precision_recall_curve(precision, recall, model_name)

	# Learning Curve
	print("Calculating Learning Curve")
	train_sizes, train_scores, test_scores = learning_curve(model, X_train, y_train.values.ravel(), cv=5,
	                                                        scoring="accuracy")
	plot_learning_curve(train_sizes, train_scores, test_scores, model_name)

	# Calculate prediction probabilities for ROC curve
	y_score = model.predict_proba(X_test)[:, 1]

	# Plot ROC curve using the function from utilities
	plot_roc_curve(y_test, y_score, model_name=model_name)

	# Plot quantity of elements in class 0 and class 1
	plot_counts(y_data, df_train)

	print("Graphs complete.")
